﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX;
using System.Drawing.Drawing2D;
using DrRobot.P2PSpeedDrive;
using DrRobot.SensorMapBuilder;


namespace DrRobot.X80Demo
{
    public partial class DrRobotX80DemoForm : Form
    {
        #region variables define

        private const short NOCONTROL = -32767;
        private const double MAXSPEED = 1.0;        //1m /s
        private const double MINSPEED = 0.15;        // 0.15m /s
        private const int MINPWM = 8000;            //the minimum PWM offset to let robot move
        private const int ENCODERHALFRANGE = 16384;
        private const int ENCODERFULLRANGE = 32767;
        private const double STARGAZEROFFSETX = 0.13;       //unit:m
        private const double STARGAZEROFFSETY = 0;
        private const double MotorHeatCurrentTH = 0.7;        //0.8A
        private const int MotorHeatAccTH = 90;
        private const double StuckCurrentTH = 0.7;       //0.7A
        private const int MaxStuckLevel = 30;
        private const int StuckLevelTH = 18;
        private const int StuckLevelUpStep = 6;
        private const int StuckLevelDownStep = 1;
        private const int StuckLevelResetTH = 10;
        private const int SetChargeTimer = 7200;        //2 hours
        private const int BatteryTemperatureLM = 1200;  //battery temperature limit
        private const int ServoNum = 2;
        //for joystick
        public static JoystickState state = new JoystickState();
        private Device applicationDevice = null;
        private bool driveCheck = false;
        private bool joyStickCA = true ;

 
        //sensor config
        private static int irNum = 0;
        private static int usNum = 0;

        //for StarGazer communication
 


        //for drawing map
        double MapRatio = 35;
        int MapX0 = 50;
        int MapY0 = 50;
        double HRes = 96;
        double VRes = 96;

        //for P2P Speed Drive
        private int P2PTaskPointNum = 0;        //p2p task total point number
        private int P2PTaskPointCnt = 0;        //p2p task point cnt
        private bool MultiPointsTask = false;       //single point p2p task
        private PointSet pointSet = new PointSet();
        private P2PTaskType _p2pTaskType = P2PTaskType.P2PTask;
        
        //motor sensor data and state
        private MotorState leftMotor = new MotorState();
        private MotorState rightMotor = new MotorState();
        private bool firstCal = true;
        private int preLeftWheelPos = 0;
        private int preRightWheelPos = 0;

        private DrRobotP2PSpeedDrive p2pDrive = new DrRobotP2PSpeedDrive();
        private DrRobotSensorMapBuilder sensorMapBuilder = new DrRobotSensorMapBuilder();
        
        private DrRobotP2PSpeedDrive.RobotPosition robotPosition = new DrRobotP2PSpeedDrive.RobotPosition();
        private DrRobotSensorMapBuilder.RobotPosition robotMapPosition = new DrRobotSensorMapBuilder.RobotPosition();
        
        
        private delegate void UpdateP2PDriveCmdDelegate(DrRobotP2PSpeedDrive.MotorControlCmd cmd);
        private delegate void UpdateMapServiceDelegate(DrRobotSensorMapBuilder.SensorMapData mapData);
        private DrRobotP2PSpeedDrive.RobotInfo robotInfo = new DrRobotP2PSpeedDrive.RobotInfo();
        private IrSensorConfigSet irSensorConfigSet = new IrSensorConfigSet();
        private DrRobotSensorMapBuilder.ConfigData sensorMapConfig = new DrRobotSensorMapBuilder.ConfigData();
        private UsSensorConfigSet usSensorConfigSet = new UsSensorConfigSet();
        private DrRobotSensorMapBuilder.SensorData sensorData = new DrRobotSensorMapBuilder.SensorData();
        private DrRobotP2PSpeedDrive.P2PServiceStatus robotP2PStatus = new DrRobotP2PSpeedDrive.P2PServiceStatus();
        private DrRobotP2PSpeedDrive.SensorMapData sendMapData = new DrRobotP2PSpeedDrive.SensorMapData();

        //for servo
        private DataSetServoConfig _servoConfig = new DataSetServoConfig();
        private ServoConfigSet[] servoConfigSet = new ServoConfigSet[12];
        private int ServoDemoStep = 0;

        #endregion

        #region form event

        public DrRobotX80DemoForm()
        {
            InitializeComponent();

            //set robot information here
            robotInfo.MotorDir = 1;     //I90 series = -1, X80 and Sputnik = 1
            robotInfo.OneCircleCount = 1200; //I90 series = 800, X80 and Sputnik = 1200
            robotInfo.WheelRadius = 0.0825;     //unit :m
            robotInfo.WheelDis = 0.26;  ////I90 series = 0.30, X80 and Sputnik = 0.26
            p2pDrive.SetRobotInfo(robotInfo);


            robotPosition.robotX = 0;
            robotPosition.robotY = 0;
            robotPosition.robotDir = 0;

            robotMapPosition.robotX = 0;
            robotMapPosition.robotY = 0;
            robotMapPosition.robotDir = 0;

            //read Ir and US sensor config file
            string directory = System.Environment.CurrentDirectory;
            try
            {
                irSensorConfigSet.ReadXml(directory + "\\IrSensorConfig.xml");
            }
            catch
            {
                MessageBox.Show("Unable to find IR Ranger Sensor configure file");
                
            }
           // irSensorConfigSet.ReadXml("c:\\IrSensorConfig.xml");
            irNum = irSensorConfigSet.DataTableIrSensor.Count;
            sensorMapConfig.IrNum = irNum;
            if (irNum <= 20)  //must less than 20 Ir sensor
            {
                for (int i = 0; i < irNum; i++)
                {
                    IrSensorConfigSet.DataTableIrSensorRow row = (IrSensorConfigSet.DataTableIrSensorRow)irSensorConfigSet.DataTableIrSensor.Rows[i];
                    sensorMapConfig.IrConfigData[i].OffsetX = double.Parse(row.OffsetX);
                    sensorMapConfig.IrConfigData[i].OffsetY = double.Parse(row.OffsetY);
                    sensorMapConfig.IrConfigData[i].Angle = double.Parse(row.Angle);
                    sensorMapConfig.IrConfigData[i].Weight = double.Parse(row.Weight);
                    sensorMapConfig.IrConfigData[i].DisTag = int.Parse(row.DisTag);
                }
            }

            //Ultrasonic config file
            try
            {
                usSensorConfigSet.ReadXml(directory + "\\UsSensorConfig.xml");
            }
            catch
            {
                MessageBox.Show("Unable to find Ultrasonic Sensor configure file");
            }
          //  usSensorConfigSet.ReadXml("c:\\UsSensorConfig.xml");
            usNum = usSensorConfigSet.DataTableUsSensor.Count;
            sensorMapConfig.UsNUm = usNum;

            if (usNum <= 6)           //not more than 6 Us sensor
            {
                for (int i = 0; i < usNum; i++)
                {
                    UsSensorConfigSet.DataTableUsSensorRow row = (UsSensorConfigSet.DataTableUsSensorRow)usSensorConfigSet.DataTableUsSensor.Rows[i];
                    sensorMapConfig.UsConfigData[i].OffsetX = double.Parse(row.OffsetX);
                    sensorMapConfig.UsConfigData[i].OffsetY = double.Parse(row.OffsetY);
                    sensorMapConfig.UsConfigData[i].Angle = double.Parse(row.Angle);
                    sensorMapConfig.UsConfigData[i].Weight = double.Parse(row.Weight);
                    sensorMapConfig.UsConfigData[i].DisTag = int.Parse(row.DisTag);
                }
            }
 

        }

        private void DrRobotX80DemoForm_Load(object sender, EventArgs e)
        {
            //connect to activeX control
            motionControl.connectRobot("motion");

            tmrImage.Enabled = true;


            //register service call back function
            p2pDrive.RegisterP2PDriveCmdCallback(UpdateP2PCmd);
            sensorMapBuilder.RegisterSensorMapCallback(UpdateSensorMap);
            
            

            //config sensor
            sensorMapBuilder.ConfigSensor(sensorMapConfig);
            //must preset positon for sensormapbuilder
            sensorMapBuilder.PreSetPosition(robotMapPosition);
            //sart send sensor information to sensor map builder service
            tmrSensorMapBuilder.Enabled = true;

            //init pointset here
            //PointSet pointSet = new PointSet();
            PointSet.PointConfigTableRow row = (PointSet.PointConfigTableRow)pointSet.PointConfigTable.NewRow();
            //set target at the 1m away with the joystick direction
            row.TargetX = -1;
            row.TargetY = 0;
            row.TargetDir = 0;

            row.ForwardSpeed = 0.3;
            row.CAEnable = false;
            row.FinalPosture = true ;
            row.Forgetable = false;
            row.MaxTurnSpeed = 35;      //35 degree
            row.NonStop = false;

            row.ReverseDrive = true;

            row.StopTime = 0;
            row.TargetTime = 200;
            row.TargetTolerance = 0.1;
            row.TargetDirTolerance = 15;        //15 degree

            pointSet.PointConfigTable.AddPointConfigTableRow(row);
            dataGridViewPointConfig.DataSource = pointSet.PointConfigTable;
            int columnCnt = dataGridViewPointConfig.ColumnCount;
            for (int i = 0; i < columnCnt; i++)
            {
                dataGridViewPointConfig.Columns[i].Width = (dataGridViewPointConfig.Width - dataGridViewPointConfig.RowHeadersWidth) / columnCnt;
            }

            txtMapRatio.Text = MapRatio.ToString();
            MapX0 = pictureBoxMap.Width / 2;
            MapY0 = pictureBoxMap.Height / 2;
            txtMapX0.Text = MapX0.ToString();
            txtMapY0.Text = MapY0.ToString();

            drawMap();
            drawRobot();
            drawPath();

            

            //init Joystick
            if (InitDirectInput())
                tmrJoyStickPoll.Start();

            //to update sensor map and robot position to p2p drive service
            tmrP2PUpdate.Enabled = true;
            

            //set servo config
            string directory = System.Environment.CurrentDirectory;
            _servoConfig.ReadXml(directory + "\\servoconfig.xml");

            int cnt = _servoConfig.ServoConfigDataTable.Count;
            if (cnt != ServoNum)
            {
                return;
            }
            else
            {
                for (int j = 0; j < ServoNum; j++)
                {
                    DataSetServoConfig.ServoConfigDataTableRow rowHead = (DataSetServoConfig.ServoConfigDataTableRow)_servoConfig.ServoConfigDataTable.Rows[j];

                    servoConfigSet[j] = new ServoConfigSet();
                    servoConfigSet[j].Ini = rowHead.Ini;
                    servoConfigSet[j].Min = rowHead.Min - 1;
                    servoConfigSet[j].Max = rowHead.Max + 1;
                    
                }
                trackBarHeadPan.Maximum = servoConfigSet[1].Max;
                trackBarHeadPan.Minimum = servoConfigSet[1].Min;
                trackBarHeadPan.Value = servoConfigSet[1].Ini;

                trackBarHeadTilt.Maximum = servoConfigSet[0].Max;
                trackBarHeadTilt.Minimum = servoConfigSet[0].Min;
                trackBarHeadTilt.Value = servoConfigSet[0].Ini;

                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)servoConfigSet[1].Ini,NOCONTROL,NOCONTROL,NOCONTROL,NOCONTROL, 1000);
            }

        }

        private void DrRobotX80DemoForm_FormClosed(object sender, FormClosedEventArgs e)
        {
       
            p2pDrive.DeregisterP2PDriveCmdCallback();
            sensorMapBuilder.DeregisterSensorMapCallback();


        }

        private void motionControl_StandardSensorEvent(object sender, EventArgs e)
        {
            sensorData.UsDis[0] = (double)motionControl.GetSensorSonar1()/100;
            sensorData.UsDis[1] = (double)motionControl.GetSensorSonar2()/100;
            sensorData.UsDis[2] = (double)motionControl.GetSensorSonar3()/100;

            lblUsDis1.Text = ((double)(sensorData.UsDis[0])).ToString() + "m";
            lblUsDis2.Text = ((double)(sensorData.UsDis[1])).ToString() + "m";
            lblUsDis3.Text = ((double)(sensorData.UsDis[2])).ToString() + "m";

            sensorData.IrDis[0] = AD2Dis(motionControl.GetSensorIRRange());

            lblIRDis1.Text = (sensorData.IrDis[0]).ToString("0.00") + "m";
            lblBoardVol.Text = ((double)(motionControl.GetSensorBatteryAD1()) / 4095.0 * 9).ToString("0.00") + "V";
            lblMotorVol.Text = ((double)(motionControl.GetSensorBatteryAD2()) / 4095.0 * 24).ToString("0.00") + "V";

            lblLeftHumanAlarm.Text = (motionControl.GetSensorHumanAlarm1()).ToString();
            lblLeftHumanMotion.Text = (motionControl.GetSensorHumanMotion1()).ToString();
            lblRightHumanAlarm.Text = (motionControl.GetSensorHumanAlarm2()).ToString();
            lblRightHumanMotion.Text = (motionControl.GetSensorHumanMotion2()).ToString();
        }

        private void motionControl_CustomSensorEvent(object sender, EventArgs e)
        {
            sensorData.IrDis[1] = AD2Dis(motionControl.GetCustomAD3());
            sensorData.IrDis[2] = AD2Dis(motionControl.GetCustomAD4());
            sensorData.IrDis[3] = AD2Dis(motionControl.GetCustomAD5());
            sensorData.IrDis[4] = AD2Dis(motionControl.GetCustomAD6());
            sensorData.IrDis[5] = AD2Dis(motionControl.GetCustomAD7());
            sensorData.IrDis[6] = AD2Dis(motionControl.GetCustomAD8());

            lblIRDis2.Text = (sensorData.IrDis[1]).ToString("0.00") + "m";
            lblIRDis3.Text = (sensorData.IrDis[2]).ToString("0.00") + "m";
            lblIRDis4.Text = (sensorData.IrDis[3]).ToString("0.00") + "m";
            lblIRDis5.Text = (sensorData.IrDis[4]).ToString("0.00") + "m";
            lblIRDis6.Text = (sensorData.IrDis[5]).ToString("0.00") + "m";
            lblIRDis7.Text = (sensorData.IrDis[6]).ToString("0.00") + "m";
            short ioinput = motionControl.GetCustomDIN();
            lblInputIO.Text = ioinput.ToString();
            
            
        }

        private void motionControl_MotorSensorEvent(object sender, EventArgs e)
        {
            leftMotor.encoderPos = motionControl.GetEncoderPulse1();
            leftMotor.encoderSpeed = (motionControl.GetEncoderSpeed1());
            leftMotor.current = (double)(motionControl.GetMotorCurrent1()) / 728.0;

            rightMotor.encoderPos = motionControl.GetEncoderPulse2();
            rightMotor.encoderSpeed = (motionControl.GetEncoderSpeed2());
            rightMotor.current = (double)(motionControl.GetMotorCurrent2()) / 728.0;

            robotPosition.LeftWheelEncoderCnt = leftMotor.encoderPos;
            robotPosition.RightWheelEncoderCnt = rightMotor.encoderPos;

            lblLeftMotorEncoderPos.Text = leftMotor.encoderPos.ToString();;
            lblRightMotorEncoderPos.Text = rightMotor.encoderPos.ToString();

            lblLeftMotorEncoderSpeed.Text = leftMotor.encoderSpeed.ToString();
            lblRightMotorEncoderSpeed.Text = rightMotor.encoderSpeed.ToString();

            lblLeftMotorCurrentFeedBack.Text = leftMotor.current.ToString("0.00") + "A";

            lblRightMotorCurrentFeedBack.Text = rightMotor.current.ToString("0.00") + "A";

            encoderGps();
            StucKHeatDetect();

        }

        
        private void btnForward_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;
            short speed = (short)hScrollBar1.Value;
            motionControl.DcMotorVelocityNonTimeCtrAll((short)(-robotInfo.MotorDir * speed), (short)(robotInfo.MotorDir * speed), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
        }

        private void btnTurnLeft_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;
            short speed = (short)hScrollBar1.Value;
            motionControl.DcMotorVelocityNonTimeCtrAll((short)(robotInfo.MotorDir * speed), (short)(robotInfo.MotorDir * speed), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
        }

        private void btnBackward_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;
            short speed = (short)hScrollBar1.Value;
            motionControl.DcMotorVelocityNonTimeCtrAll((short)(robotInfo.MotorDir * speed), (short)(-robotInfo.MotorDir * speed), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;


            motionControl.DisableDcMotor(0);
            motionControl.DisableDcMotor(1);
        }

        private void btnTurnRight_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;
            short speed = (short)hScrollBar1.Value;
            motionControl.DcMotorVelocityNonTimeCtrAll((short)(-robotInfo.MotorDir * speed), (short)(-robotInfo.MotorDir * speed), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
        }

        private void btnSetDrivePower_Click(object sender, EventArgs e)
        {
            if (btnSetDrivePower.Text == "Go")
            {
                short cmd1 = 0;
                short cmd2 = 0;

                chkEnableJoystick.Checked = false;
                try
                {
                    cmd1 = short.Parse(txtLeftDrivePower.Text);
                    cmd2 = short.Parse(txtRightDrivePower.Text);
                }
                catch
                {
                    cmd1 = 16384;
                    cmd2 = 16384;
                }
                txtRightDrivePower.Text = cmd2.ToString();
                txtLeftDrivePower.Text = cmd1.ToString();


                motionControl.DcMotorPwmNonTimeCtrAll(cmd1, cmd2, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
                btnSetDrivePower.Text = "Stop";
            }
            else
            {
                btnSetDrivePower.Text = "Go";
                motionControl.DcMotorPwmNonTimeCtrAll(16384, 16384, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
            }
        }

        private void btnSetDriveSpeed_Click(object sender, EventArgs e)
        {
            if (btnSetDriveSpeed.Text == "Go")
            {
                short cmd1 = 0;
                short cmd2 = 0;

                chkEnableJoystick.Checked = false;
                try
                {
                    cmd1 = short.Parse(txtLeftSpeed.Text);
                    cmd2 = short.Parse(txtRightSpeed.Text);
                }
                catch
                {
                    cmd1 = 0;
                    cmd2 = 0;
                }
                txtLeftSpeed.Text = cmd1.ToString();
                txtRightSpeed.Text = cmd2.ToString();
                motionControl.DcMotorVelocityNonTimeCtrAll(cmd1, cmd2, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
                btnSetDriveSpeed.Text = "Stop";
            }
            else
            {
                btnSetDriveSpeed.Text = "Go";
                motionControl.DcMotorVelocityNonTimeCtrAll(0, 0, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
            }
        }

        private void btnRotateDegree_Click(object sender, EventArgs e)
        {
            chkEnableJoystick.Checked = false;
            double degree = 0;
            try
            {
                degree = double.Parse(txtRotateDegree.Text);
            }
            catch
            {
                degree = 0;
                txtRotateDegree.Text ="0";
            }

            short time = 0;
            try
            {
                time = short.Parse(txtRotateDegreePower.Text);
            }
            catch
            {
                time = 1000;
                txtRotateDegreePower.Text = "1000";
            }

            double Dis = robotInfo.WheelDis / 2 * (degree / 180.0 * Math.PI);
            int diffEncoder = (int)(Dis / (2 * Math.PI * robotInfo.WheelRadius) * robotInfo.OneCircleCount);

            int LeftCmd = int.Parse(lblLeftMotorEncoderPos.Text) + robotInfo.MotorDir * diffEncoder;
            if (LeftCmd < 0)
            {
                LeftCmd = 32767 + LeftCmd;
            }
            else if (LeftCmd > 32767)
            {
                LeftCmd = LeftCmd - 32767;
            }

            int RightCmd = int.Parse(lblRightMotorEncoderPos.Text) + robotInfo.MotorDir * diffEncoder;
            if (RightCmd < 0)
            {
                RightCmd = 32767 + RightCmd;
            }
            else if (RightCmd > 32767)
            {
                RightCmd = RightCmd - 32767;
            }
            motionControl.SetDcMotorPositionControlPID(0, 1000, 5, 10000);
            motionControl.SetDcMotorPositionControlPID(1, 1000, 5, 10000);
            motionControl.DcMotorPositionTimeCtrAll((short)(LeftCmd), (short)(RightCmd), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, time);
        }

        private void btnDriveDistance_Click(object sender, EventArgs e)
        {
            double Dis = 1.0;        //1m
            try
            {
                Dis = double.Parse(txtDriveDistance.Text);
            }
            catch
            {
                Dis = 0;
                txtDriveDistance.Text = "0";
            }
            short time = 0;
            try
            {
                time = short.Parse(txtDriveDistancePower.Text);
            }
            catch
            {
                time = 1000;
                txtDriveDistancePower.Text = "1000";
            }


            int diffEncoder = (int)(Dis / (2 * Math.PI * robotInfo.WheelRadius) * robotInfo.OneCircleCount);

            int LeftCmd = int.Parse(lblLeftMotorEncoderPos.Text) - robotInfo.MotorDir * diffEncoder;
            if (LeftCmd < 0)
            {
                LeftCmd = 32767 + LeftCmd;
            }
            else if (LeftCmd > 32767)
            {
                LeftCmd = LeftCmd - 32767;
            }

            int RightCmd = int.Parse(lblRightMotorEncoderPos.Text) + robotInfo.MotorDir * diffEncoder;
            if (RightCmd < 0)
            {
                RightCmd = 32767 + RightCmd;
            }
            else if (RightCmd > 32767)
            {
                RightCmd = RightCmd - 32767;
            }

            motionControl.DcMotorPositionTimeCtrAll((short)LeftCmd, (short)RightCmd, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, time);

        }


   
        private void tmrJoyStickPoll_Tick(object sender, EventArgs e)
        {
            GetData();
        }

        private void tmrSensorMapBuilder_Tick(object sender, EventArgs e)
        {
            robotMapPosition.robotX = robotPosition.robotX;
            robotMapPosition.robotY = robotPosition.robotY;
            robotMapPosition.robotDir = robotPosition.robotDir;

            //for test
            /*
            for (int i = 0; i < 7; i++)
            {
                sensorData.IrDis[i] = 0.81;
            }
            for (int i = 0; i < 6; i++)
            {
                sensorData.UsDis[i] = 2.5;
            }

            sensorData.IrDis[1] = 0.40;
            */
            sensorMapBuilder.UpdateSensorInfo(sensorData, robotMapPosition);
        }
    
    
        private void btnOpenPathFile_Click(object sender, EventArgs e)
        {
            openFileDialog1.FileName = "";
            openFileDialog1.Filter = "Path XML Files (*.xml)|*.xml";
            openFileDialog1.Title = "Path XML Type File";
            DialogResult result = openFileDialog1.ShowDialog();
            if (result == DialogResult.OK)
            {
                try
                {
                    txtOpenPathFile.Text = openFileDialog1.FileName;
                    pointSet.Clear();
                    pointSet.ReadXml(openFileDialog1.FileName);
                    dataGridViewPointConfig.DataSource = pointSet.PointConfigTable;
                    drawPath();

                }
                catch
                {
                    MessageBox.Show("Error opening file", "File Error",
                                     MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
                return;
            }
            else
            {

                if (result == DialogResult.Cancel)
                {
                    return;
                }
            }
        }

        private void txtMapRatio_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)13)
            {
                if (MapRatio != double.Parse(txtMapRatio.Text))
                {
                    try
                    {
                        MapRatio = double.Parse(txtMapRatio.Text);
                    }
                    catch
                    {
                        MapRatio = 35.0;
                    }
                    drawMap();
                    drawRobot();
                    drawPath();
                }
            }
        }

        private void txtMapX0_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)13)
            {
                if (MapX0 != int.Parse(txtMapX0.Text))
                {
                    try
                    {
                        MapX0 = int.Parse(txtMapX0.Text);
                    }
                    catch
                    {
                        MapX0 = pictureBoxMap.Width / 2;

                    }
                    drawMap();
                    drawRobot();
                    drawPath();
                }
            }
        }

        private void txtMapY0_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)13)
            {
                if (MapY0 != int.Parse(txtMapY0.Text))
                {
                    try
                    {
                        MapY0 = int.Parse(txtMapY0.Text);
                    }
                    catch
                    {

                        MapY0 = pictureBoxMap.Height / 2;
                    }
                    drawMap();
                    drawRobot();
                    drawPath();
                }
            }
        }

        private void pictureBoxMap_MouseClick(object sender, MouseEventArgs e)
        {
            //set the point by mouse
            if (chkMousePointToMove.Checked)
            {
                pointSet.Clear();
                PointSet.PointConfigTableRow row = (PointSet.PointConfigTableRow)pointSet.PointConfigTable.NewRow();
                //set target at the 1m away with the joystick direction
                double x = (double)(MapY0 - e.Y) / VRes * 0.0254 * MapRatio;
                double y = (double)(MapX0 - e.X) / HRes * 0.0254 * MapRatio;
                row.TargetX = double.Parse(x.ToString("0.00"));
                row.TargetY = double.Parse(y.ToString("0.00"));
                row.TargetDir = 0;

                row.ForwardSpeed = 0.3;
                row.CAEnable = true;
                row.FinalPosture = false;
                row.Forgetable = true;
                row.MaxTurnSpeed = 35;      //35 degree
                row.NonStop = true;

                row.ReverseDrive = false;

                row.StopTime = 0;
                row.TargetTime = 200;
                row.TargetTolerance = 0.1;
                row.TargetDirTolerance = 15;        //15 degree

                pointSet.PointConfigTable.AddPointConfigTableRow(row);
                dataGridViewPointConfig.DataSource = pointSet.PointConfigTable;
                drawPath();
            }
        }

        private void btnP2PGo_Click(object sender, EventArgs e)
        {
            if (_p2pTaskType == P2PTaskType.WanderTask)
            {
                //generate next target point
                wanderingNextPoint();
                tmrControl.Enabled = true;

            }
            else
            {
                P2PTaskPointNum = pointSet.PointConfigTable.Count;
                P2PTaskPointCnt = 0;

                SendP2PGoCmd(P2PTaskPointCnt);

                //for test
                
                
                if (P2PTaskPointNum > 1)
                {
                    //multi points to move, set flag here
                    MultiPointsTask = true;
                    tmrControl.Enabled = true;

                }
                else
                {
                    MultiPointsTask = false;
                }

            }
        }
 
        private void btnP2PStop_Click(object sender, EventArgs e)
        {
            MultiPointsTask = false;
            SendP2PStopCmd();
            tmrControl.Enabled = false;
            motionControl.DisableDcMotor(0);
            motionControl.DisableDcMotor(1);

        }

        private void tmrControl_Tick(object sender, EventArgs e)
        {
            if (_p2pTaskType == P2PTaskType.WanderTask)
            {
                //generate the next point
                wanderingNextPoint();
            }
            else
            {
                if (_p2pTaskType == P2PTaskType.ChargeTask)
                {
                  
                }

                if ((robotP2PStatus == DrRobotP2PSpeedDrive.P2PServiceStatus.P2POver))
                {
                    if (MultiPointsTask)
                    {
                        P2PTaskPointCnt++;

                        if (P2PTaskPointCnt < P2PTaskPointNum)
                        {
                            SendP2PGoCmd(P2PTaskPointCnt);
                            
                        }
                        else
                        {
                            if (_p2pTaskType == P2PTaskType.ChargeTask)
                            {
                                
                                
                            }
                            else if (_p2pTaskType == P2PTaskType.PatrolTask)
                            {
                                P2PTaskPointCnt = 0;
                                SendP2PGoCmd(P2PTaskPointCnt);
                                
                            }
                            else
                            {
                                SendP2PStopCmd();

                                MultiPointsTask = false;
                            }
                        }
                    }
                }
            }
        }

        private void comboBoxP2PTask_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (comboBoxP2PTask.Text == "P2P Task")
                _p2pTaskType = P2PTaskType.P2PTask;
            else if (comboBoxP2PTask.Text == "Patrol Task")
                _p2pTaskType = P2PTaskType.PatrolTask;
            else if (comboBoxP2PTask.Text == "Wander Task")
                _p2pTaskType = P2PTaskType.WanderTask;
        }

        

        private void tmrP2PUpdate_Tick(object sender, EventArgs e)
        {
            p2pDrive.UpdateSensorMap_Position(robotPosition, sendMapData);
            drawRobot();

        }

        private void dataGridViewPointConfig_CellEndEdit(object sender, DataGridViewCellEventArgs e)
        {
            drawPath();
        }

        private void btnSetOutputIO_Click(object sender, EventArgs e)
        {
            short ioout = 0;
            try
            {
                ioout = short.Parse(txtOutputIO.Text);
            }
            catch
            {
                ioout = 0;
            }
            txtOutputIO.Text = ioout.ToString();
            motionControl.SetCustomDOUT((short)ioout);
        }

        private void btnPlayMusic_Click(object sender, EventArgs e)
        {
            if (btnPlayMusic.Text == "Play Music")
            {

                openFileDialog2.InitialDirectory = ".\\";
                openFileDialog2.FileName = "";
                openFileDialog2.Filter = "Wave Files (*.wav)|*.wav";
                openFileDialog2.Title = "Type File";
                //openFileDialog1.ShowDialog();

                if (openFileDialog2.ShowDialog() != DialogResult.OK)
                    return;
                else
                {
                    txtMusicFilePath.Text = openFileDialog2.FileName;
                    motionControl.PlayAudioFile(openFileDialog2.FileName);
                    btnPlayMusic.Text = "Stop Music";

                }

            }
            else
            {
                motionControl.StopAudioPlay();
                btnPlayMusic.Text = "Play Music";
            }
        }

        private void btnSaveImage_Click(object sender, EventArgs e)
        {
            motionControl.SavePhotoAsBMP(Environment.CurrentDirectory + "photo.bmp");
        }

        private void tmrImage_Tick(object sender, EventArgs e)
        {
            motionControl.TakePhoto();
        }

        private void btnHeadReset_Click(object sender, EventArgs e)
        {
            trackBarHeadPan.Value = servoConfigSet[1].Ini;
            trackBarHeadTilt.Value = servoConfigSet[0].Ini;
            motionControl.ServoTimeCtrAll((short)trackBarHeadTilt.Value, (short)trackBarHeadPan.Value, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 2000);
        }

        private void btnHeadDemo_Click(object sender, EventArgs e)
        {
            ServoDemoStep = 0;

            motionControl.ServoTimeCtrAll((short)trackBarHeadTilt.Value, (short)trackBarHeadPan.Value, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 2000);
            tmrServoDemo.Enabled = true;
        }

        private void tmrServoDemo_Tick(object sender, EventArgs e)
        {
            ServoDemoStep++;
            if (ServoDemoStep == 1)
                motionControl.ServoTimeCtrAll((short)(servoConfigSet[0].Ini + 500), (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 2)
                motionControl.ServoTimeCtrAll((short)(servoConfigSet[0].Ini - 500), (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 3)
                motionControl.ServoTimeCtrAll((short)(servoConfigSet[0].Ini + 500), (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 4)
                motionControl.ServoTimeCtrAll((short)(servoConfigSet[0].Ini - 500), (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 5)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 6)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)(servoConfigSet[1].Ini + 500), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 7)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)(servoConfigSet[1].Ini - 500), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 8)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)(servoConfigSet[1].Ini + 500), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 9)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)(servoConfigSet[1].Ini - 500), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 10)
                motionControl.ServoTimeCtrAll((short)servoConfigSet[0].Ini, (short)servoConfigSet[1].Ini, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, 1500);
            else if (ServoDemoStep == 11)
            {
                ServoDemoStep = -1;
                tmrServoDemo.Enabled = false;
            }

        }
        private void btnPosReset_Click(object sender, EventArgs e)
        {
            robotPosition.robotX = 0;
            robotPosition.robotY = 0;
            robotPosition.robotDir = 0;

            robotMapPosition.robotX = 0;
            robotMapPosition.robotY = 0;
            robotMapPosition.robotDir = 0;
            firstCal = true;

        }

        private void trackBarHeadPan_Scroll(object sender, EventArgs e)
        {
            int time = 0;
            try
            {
                time = int.Parse(txtHeadPanTime.Text);
            }
            catch
            {
                time = 2000;
                txtHeadPanTime.Text = time.ToString();
            }

            motionControl.ServoTimeCtr(1, (short)(trackBarHeadPan.Maximum + trackBarHeadPan .Minimum  -trackBarHeadPan.Value), (short)time);

        }

        private void trackBarHeadTilt_Scroll(object sender, EventArgs e)
        {
            int time = 0;
            try
            {
                time = int.Parse(txtHeadTiltTime.Text);
            }
            catch
            {
                time = 2000;
                txtHeadTiltTime.Text = time.ToString();
            }

            motionControl.ServoTimeCtr(0, (short)trackBarHeadTilt.Value, (short)time);
        }

        #endregion

        #region joystick control
        private bool InitDirectInput()
        {
            // Enumerate joysticks in the system.
            foreach (DeviceInstance instance in Manager.GetDevices(DeviceClass.GameControl, EnumDevicesFlags.AttachedOnly))
            {
                // Create the device.  Just pick the first one
                applicationDevice = new Device(instance.InstanceGuid);
                break;
            }

            if (null == applicationDevice)
            {
                MessageBox.Show("Unable to create a joystick device.", "No joystick found");
                return false;
            }

            // Set the data format to the c_dfDIJoystick pre-defined format.
            applicationDevice.SetDataFormat(DeviceDataFormat.Joystick);
            // Set the cooperative level for the device.
            applicationDevice.SetCooperativeLevel(this, CooperativeLevelFlags.Exclusive | CooperativeLevelFlags.Foreground);
            // Enumerate all the objects on the device.
            foreach (DeviceObjectInstance d in applicationDevice.Objects)
            {
                // For axes that are returned, set the DIPROP_RANGE property for the
                // enumerated axis in order to scale min/max values.

                if ((0 != (d.ObjectId & (int)DeviceObjectTypeFlags.Axis)))
                {
                    // Set the range for the axis.
                    applicationDevice.Properties.SetRange(ParameterHow.ById, d.ObjectId, new InputRange(0, +10000));
                }
                // Update the controls to reflect what
                // objects the device supports.
                //  UpdateControls(d);
            }
            return true;
        }
        private void GetData()
        {
            // Make sure there is a valid device.
            if (null == applicationDevice)
                return;

            try
            {
                // Poll the device for info.
                applicationDevice.Poll();
            }
            catch (InputException inputex)
            {
                if ((inputex is NotAcquiredException) || (inputex is InputLostException))
                {
                    // Check to see if either the app
                    // needs to acquire the device, or
                    // if the app lost the device to another
                    // process.
                    try
                    {
                        // Acquire the device.
                        applicationDevice.Acquire();
                    }
                    catch (InputException)
                    {
                        // Failed to acquire the device.
                        // This could be because the app
                        // doesn't have focus.
                        return;
                    }
                }

            } //catch(InputException inputex)

            // Get the state of the device.
            try { state = applicationDevice.CurrentJoystickState; }
            // Catch any exceptions. None will be handled here, 
            // any device re-aquisition will be handled above.  
            catch (InputException)
            {
                return;
            }

            UpdateControl();

        }
        private void UpdateControl()
        {
            byte[] buttons = state.GetButtons();

            if ((buttons[0] & 0x80) != 0)
                driveCheck = false;
            else
                driveCheck = true;

            if ((buttons[1] & 0x80) != 0)
                joyStickCA = false ;
            else
                joyStickCA = true ;

            int x = state.Y;
            int y = state.X;

            if (chkEnableJoystick.Checked)
            {
                RobotDrive(x, y);
            }

        }

        private void RobotDrive(int gasLevel, int SteerLevel)
        {
            int x = SteerLevel - 5000;
            int y = -(gasLevel - 5000);
            double left = 0;
            double right = 0;           

            DrawJoystick(x, y);
            
            //Steeringwheel2PWM();
            if (!chkJoyCADrive.Checked)
            {
                //here is direct PWM drive, no collision avoidance
                if (driveCheck)
                {


                    //this is the raw length of the vector
                    double magnitude = Math.Sqrt(y * y + x * x);

                    //angle of the vector
                    double theta = Math.Atan2(y, x);

                    //this is the maximum magnitude for a given angle
                    double maxMag;
                    if (Math.Abs(x) > Math.Abs(y))
                        maxMag = Math.Abs(5000 / Math.Cos(theta));
                    else
                        maxMag = Math.Abs(5000 / Math.Sin(theta));

                    //a scaled down magnitude according to above
                    double scaledMag = magnitude * 1000 / maxMag;

                    //decompose the vector into motor components
                    left = scaledMag * Math.Sin(theta) + scaledMag * Math.Cos(theta) / 3;
                    right = scaledMag * Math.Sin(theta) - scaledMag * Math.Cos(theta) / 3;
                }
                else
                {
                    left = right = 0;
                }

                //cap at 1000
                left = Math.Min(left, 1000);
                right = Math.Min(right, 1000);
                left = Math.Max(left, -1000);
                right = Math.Max(right, -1000);

                //throttle control
                left *= (double)throttle.Value / 100.0;
                right *= (double)throttle.Value / 100.0;
                
                

                double LeftWheelPower = left / 1000 * (16384 - MINPWM);
                double RightWheelPower = right / 1000 * (16384 - MINPWM);

                if (Math.Abs(LeftWheelPower) > 100)
                    LeftWheelPower = 16384 - robotInfo.MotorDir * (Math.Sign(LeftWheelPower) * MINPWM + LeftWheelPower);
                else
                    LeftWheelPower = 16384;

                if (LeftWheelPower > 32767)
                    LeftWheelPower = 32767;
                else if (LeftWheelPower < 0)
                    LeftWheelPower = 0;

                if (Math.Abs(RightWheelPower) > 100)
                    RightWheelPower = 16384 + robotInfo.MotorDir * (Math.Sign(RightWheelPower) * MINPWM + RightWheelPower);
                else
                    RightWheelPower = 16384;

                if (RightWheelPower > 32767)
                    RightWheelPower = 32767;
                else if (RightWheelPower < 0)
                    RightWheelPower = 0;

                lblLeftWheelJoystick.Text = LeftWheelPower.ToString("0");
                lblRightWheelJoystick.Text = RightWheelPower.ToString("0");

                motionControl.DcMotorPwmNonTimeCtrAll((short)LeftWheelPower, (short)RightWheelPower, NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
            }
            else
            {
                //here is the Joystick with CA(collision avoidance) drive
                //we need use P2P Speed drive
                //first we need to do some transform
                // Joystick X---> Map -Y   , Joystick Y --- > X
                double Xm = y;
                double Ym = -x;

                double thelta = Math.Atan2(Ym, Xm);

                PointSet _joystickset = new PointSet();
                PointSet.PointConfigTableRow row = (PointSet.PointConfigTableRow)_joystickset.PointConfigTable.NewRow();
                //if the joystick at the center position, will stop robot
                if ((Math.Abs(gasLevel - 5000) < 300) && (Math.Abs(SteerLevel - 5000) < 300))
                {//stop the robot
                    row.TargetX = robotMapPosition.robotX;
                    row.TargetY = robotPosition.robotY;
                    row.TargetDir = Double.Parse((robotPosition.robotDir * 180 / Math.PI).ToString("0.00"));
                }
                else
                {


                    //set target at the 1m away with the joystick direction
                    row.TargetX = Double.Parse((robotMapPosition.robotX + Math.Cos(thelta)).ToString("0.00"));
                    row.TargetY = Double.Parse((robotPosition.robotY + Math.Sin(thelta)).ToString("0.00"));
                    row.TargetDir = Double.Parse((robotPosition.robotDir * 180 / Math.PI).ToString("0.00"));
                }
                row.ForwardSpeed = MAXSPEED * Math.Sqrt(Xm * Xm + Ym * Ym) / 5000;
                row.ForwardSpeed = Math.Min(MAXSPEED, row.ForwardSpeed);
                row.ForwardSpeed = Math.Max(MINSPEED, row.ForwardSpeed);

                if (joyStickCA)
                    row.CAEnable = true;
                else
                    row.CAEnable = false;

                row.FinalPosture = false;
                row.Forgetable = true;
                row.MaxTurnSpeed = 35;      //35 degree
                row.NonStop = true;

                if (Xm < 0)
                    row.ReverseDrive = true;
                else
                    row.ReverseDrive = false;

                row.StopTime = 0;
                row.TargetTime = 200;
                row.TargetTolerance = 0.1;
                row.TargetDirTolerance = 15;        //15 degree

                _joystickset.PointConfigTable.AddPointConfigTableRow(row);
                dataGridViewPointConfig.DataSource = _joystickset.PointConfigTable;


                if (driveCheck)
                {
                    SendP2PGoCmd(0);
                    MultiPointsTask = false;
                }
                else
                {
                    if (robotP2PStatus != DrRobotP2PSpeedDrive.P2PServiceStatus.P2POver)
                        SendP2PStopCmd();
                }

            }

        }
        private void DrawJoystick(int x, int y)
        {
            Bitmap bmp = new Bitmap(picJoystick.Width, picJoystick.Height);
            Graphics g = Graphics.FromImage(bmp);

            int width = bmp.Width - 1;
            int height = bmp.Height - 1;


            g.Clear(Color.Transparent);
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            GraphicsPath path = new GraphicsPath();
            path.AddEllipse(0, 0, width, height);

            PathGradientBrush pathBrush = new PathGradientBrush(path);
            pathBrush.CenterPoint = new PointF(width / 3f, height / 3f);
            pathBrush.CenterColor = Color.White;
            pathBrush.SurroundColors = new Color[] { Color.LightGray };

            g.FillPath(pathBrush, path);
            g.DrawPath(Pens.Black, path);

            int partial = y * height / 5000/2;
            if (partial > 0)
            {
                g.DrawArc(Pens.Black,
                    0,
                    height / 2 - partial,
                    width,
                    2 * partial,
                    180,
                    180);
            }
            else if (partial == 0)
            {
                g.DrawLine(Pens.Black, 0, height / 2, width, height / 2);
            }
            else
            {
                g.DrawArc(Pens.Black,
                    0,
                    height / 2 + partial,
                    width,
                    -2 * partial,
                    0,
                    180);
            }

            partial = x * width / 5000/2;
            if (partial > 0)
            {
                g.DrawArc(Pens.Black,
                    width / 2 - partial,
                    0,
                    2 * partial,
                    height,
                    270,
                    180);
            }
            else if (partial == 0)
            {
                g.DrawLine(Pens.Black, width / 2, 0, width / 2, height);
            }
            else
            {
                g.DrawArc(Pens.Black,
                    width / 2 + partial,
                    0,
                    -2 * partial,
                    height,
                    90,
                    180);
            }

            picJoystick.Image = bmp;
        }
        #endregion

        #region DLL Service call back function
        //p2pDrive call back
        private void UpdateP2PCmd(DrRobotP2PSpeedDrive.MotorControlCmd cmd)
        {
            Invoke(new UpdateP2PDriveCmdDelegate(ExeP2PCmd), cmd);
        }
        private void ExeP2PCmd(DrRobotP2PSpeedDrive.MotorControlCmd cmd)
        {
            //exe p2p Comd Here
            lblP2PTest.Text = (cmd.TestVar/Math.PI * 180).ToString("0.00");

            lblP2PStatus.Text = cmd.P2PDriveStatus.ToString();
            robotP2PStatus = cmd.P2PDriveStatus;

           
            if (robotP2PStatus == DrRobotP2PSpeedDrive.P2PServiceStatus.P2POver)
            {

                motionControl.DisableDcMotor(0);
                motionControl.DisableDcMotor(1);
                return;
            }


            //always exe the wheel control cmd, 0 means stop robot

            if (checkBoxMotorProtect.Checked)
            {
                if ((leftMotor.heatUpProtect) || (rightMotor.heatUpProtect))
                {
                    //heat up, need stop
                    motionControl.DisableDcMotor(0);
                    motionControl.DisableDcMotor(1);
                    return;
                }
                else if (((leftMotor.stuckProtect) || (rightMotor.stuckProtect)) && (cmd.P2PDriveMethod == DrRobotP2PSpeedDrive.P2PDriveMethod.SpeedControl))
                {
                    cmd.LeftWheelCmd = -cmd.LeftWheelCmd;
                    cmd.RightWheelCmd = -cmd.RightWheelCmd;
                }

            }

            lblP2PCmd.Text = cmd.P2PDriveMethod.ToString();
            lblLeftWheelCmd.Text = cmd.LeftWheelCmd.ToString();
            lblRightWheelCmd.Text = cmd.RightWheelCmd.ToString();
            lblP2PCmdTime.Text = cmd.RunTime.ToString();


            if (cmd.P2PDriveMethod == DrRobotP2PSpeedDrive.P2PDriveMethod.SpeedControl)
                motionControl.DcMotorVelocityNonTimeCtrAll(short.Parse(lblLeftWheelCmd.Text), short.Parse(lblRightWheelCmd.Text), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL);
            else if (cmd.P2PDriveMethod == DrRobotP2PSpeedDrive.P2PDriveMethod.PositionControl)
                motionControl.DcMotorPositionTimeCtrAll(short.Parse(lblLeftWheelCmd.Text), short.Parse(lblRightWheelCmd.Text), NOCONTROL, NOCONTROL, NOCONTROL, NOCONTROL, short.Parse(lblP2PCmdTime.Text));
            



        }

        private void UpdateSensorMap(DrRobotSensorMapBuilder.SensorMapData mapData)
        {
            Invoke(new UpdateMapServiceDelegate(UpdateMap), mapData);
        }
        private void UpdateMap(DrRobotSensorMapBuilder.SensorMapData data)
        {
            //display polar_bin_map
            Bitmap bmp = new Bitmap(picSensorMap.Width, picSensorMap.Height);
            Graphics g = Graphics.FromImage(bmp);

            int width = bmp.Width - 1;
            int height = bmp.Height - 1;


            g.Clear(Color.Black);       //background is black
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            GraphicsPath path = new GraphicsPath();
            path.AddEllipse(0, 0, width, height);

            SolidBrush pathBrush = new SolidBrush(Color.Green);
            Pen drawPen = new Pen(pathBrush);
            drawPen.Color = Color.Yellow;


            int sectionNum = data.Polar_Bin_Map.Length;

            for (int i = 0; i < sectionNum; i++)
            {
                if (data.Polar_Bin_Map[i] != 0)
                {
                    g.FillPie(pathBrush, 0, 0, width, height, (sectionNum - i - 1) * 10 + 90, 10);  ////i * 10, 10
                }
            }

            int r = 10;

            g.DrawEllipse(drawPen, width / 2 - r, height / 2 - r, 2 * r, 2 * r);
            drawPen.Color = Color.Red;
            drawPen.Width = 2;
            int x = (int)(width / 2 - (double)(r + 5) * Math.Sin(robotPosition.robotDir));
            int y = (int)(height / 2 - (double)(r + 5) * Math.Cos(robotPosition.robotDir));

            g.DrawLine(drawPen, width / 2, height / 2, x, y);

            picSensorMap.Image = bmp;

            
            sendMapData.Polar_Bin_Map = data.Polar_Bin_Map;
            sendMapData.Polar_Value_Map = data.Polar_Value_Map;
            p2pDrive.UpdateSensorMap_Position(robotPosition, sendMapData);
        }

     
        #endregion

        #region general function
        public class ServoConfigSet
        {
            public int Min = 500;
            public int Ini = 1500;
            public int Max = 2500;
        }

        public enum P2PTaskType : int
        {
            P2PTask = 1,
            PatrolTask = 2,
            ChargeTask = 3,
            WanderTask = 4,
        }

        public class MotorState
        {
            public int encoderPos = 0;
            public int encoderSpeed = 0;
            public double current = 0;
            public double heatUpAcc = 0;
            public double stuckSpeedRatio = 0;
            public int stuckLevel = 0;
            public bool heatUpProtect = false;
            public bool stuckProtect = false;
        }

        private void encoderGps()
        {
            int leftWheelPos = robotPosition.LeftWheelEncoderCnt;
            int rightWheelPos = robotPosition.RightWheelEncoderCnt;
            double deltaLeft = 0;
            double deltaRight = 0;
            double deltaAngle = 0;

            if (firstCal == false)
            {
                //assuming no motion over 1/2 wheel cycle in every interval
                //Left Wheel travel distnace
                deltaLeft = (leftWheelPos - preLeftWheelPos);
                if (Math.Abs(deltaLeft) > ENCODERHALFRANGE)
                {
                    if (leftWheelPos < preLeftWheelPos)
                        deltaLeft = ENCODERFULLRANGE - preLeftWheelPos + leftWheelPos; //
                    else
                        deltaLeft = leftWheelPos - preLeftWheelPos - ENCODERFULLRANGE;

                }

                preLeftWheelPos = leftWheelPos;
                //change to distance        
                deltaLeft = deltaLeft / robotInfo.OneCircleCount * 2 * Math.PI;   //angle,unit radian
                deltaLeft = -robotInfo.MotorDir * deltaLeft * robotInfo.WheelRadius;

                //Right Wheel travel distnace
                deltaRight = rightWheelPos - preRightWheelPos;
                if (Math.Abs(deltaRight) > ENCODERHALFRANGE)
                {
                    if (rightWheelPos < preRightWheelPos)
                        deltaRight = ENCODERFULLRANGE - preRightWheelPos + rightWheelPos;
                    else
                        deltaRight = rightWheelPos - preRightWheelPos - ENCODERFULLRANGE;

                }

                preRightWheelPos = rightWheelPos;
                //change to distance
                deltaRight = deltaRight / robotInfo.OneCircleCount * 2 * Math.PI;  //angle,unit radian
                deltaRight = robotInfo.MotorDir * deltaRight * robotInfo.WheelRadius;

                //Update the Estimated Robot Position

                deltaAngle = (deltaRight - deltaLeft) / robotInfo.WheelDis;
                //to -2PI ~ +2Pi
                double tempAngle = Angle2PI(robotPosition.robotDir + deltaAngle);
                // to -PI ~ +PI
                tempAngle = AnglePI(tempAngle);
                //output estimate here
                robotPosition.robotDir = tempAngle;
                robotPosition.robotX = robotPosition.robotX + 0.5 * (deltaRight + deltaLeft) * Math.Cos(robotPosition.robotDir);
                robotPosition.robotY = robotPosition.robotY + 0.5 * (deltaRight + deltaLeft) * Math.Sin(robotPosition.robotDir);
                
                lblEncoderGPS.Text = "(  " + robotPosition.robotX.ToString("0.00") + ",  " + robotPosition.robotY.ToString("0.00") + ",  "
                            + (robotPosition.robotDir * 180 / Math.PI).ToString("0.00") + "  )";
                
            }
            else
            {
                preLeftWheelPos = leftWheelPos;
                preRightWheelPos = rightWheelPos;
                firstCal = false;

            }
        }

        private double Angle2PI(double tempAngle)
        {
            int numberPI;
            numberPI = (int)(Math.Abs(tempAngle) / (Math.PI * 2));
            numberPI = ((Math.Abs(tempAngle) - numberPI * (Math.PI * 2)) >= 0 ? numberPI : numberPI - 1);
            tempAngle = Math.Sign(tempAngle) * (Math.Abs(tempAngle) - numberPI * (Math.PI * 2));
            return tempAngle;
        }
        private double AnglePI(double tempAngle)
        {
            if (tempAngle > Math.PI)
                tempAngle = tempAngle - (Math.PI * 2);

            if (tempAngle < -Math.PI)
                tempAngle = tempAngle + (Math.PI * 2);

            return tempAngle;
        }

        ///<summary>
        ///draw a map here
        ///</sumary>

        private void drawMap()
        {
            //display polar_bin_map
            Bitmap bmp = new Bitmap(pictureBoxMap.Width, pictureBoxMap.Height);
            Graphics g = Graphics.FromImage(bmp);

            int width = bmp.Width - 1;
            int height = bmp.Height - 1;


            g.Clear(Color.Black);       //background is black
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            GraphicsPath path = new GraphicsPath();
            SolidBrush pathBrush = new SolidBrush(Color.Green);
            Pen drawPen = new Pen(pathBrush);
            drawPen.Color = Color.Green;

            for (int i = 0; i < width; i = i + 10)
            {
                //draw grid
                g.DrawLine(drawPen, i, 0, i, height);
            }


            for (int i = 0; i < height; i = i + 10)
            {
                //draw grid
                g.DrawLine(drawPen, 0, i, width, i);
            }


            HRes = (double)g.DpiX;
            VRes = (double)g.DpiY;
            //show the zero point on the Map
            pathBrush.Color = Color.Red;
            g.FillEllipse(pathBrush, MapX0, MapY0, 4, 4);
            pictureBoxMap.BackgroundImage = bmp;
        }

        private void drawPath()
        {
            //display polar_bin_map
            Bitmap bmp = new Bitmap(pictureBoxMap.Width, pictureBoxMap.Height);
            Graphics g = Graphics.FromImage(bmp);

            int width = bmp.Width - 1;
            int height = bmp.Height - 1;


            g.Clear(Color.Transparent);       //background is black
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            GraphicsPath path = new GraphicsPath();
            SolidBrush pathBrush = new SolidBrush(Color.White);
            Pen drawPen = new Pen(pathBrush);
            drawPen.Color = Color.Green;
            int i = pointSet.PointConfigTable.Count;
            if (i >= 1)
            {
                //have some set point
                for (int j = 0; j < i; j++)
                {
                    PointSet.PointConfigTableRow row = (PointSet.PointConfigTableRow)pointSet.PointConfigTable.Rows[j];
                    double x = row.TargetX;
                    double y = row.TargetY;
                    int MapY = MapY0 - (int)(x / MapRatio * (double)VRes / 0.0254);
                    int MapX = MapX0 - (int)(y / MapRatio * (double)HRes / 0.0254);
                    g.FillEllipse(pathBrush, MapX - 2, MapY - 2, 4, 4);
                }
            }



            pictureBoxMap.Image = bmp;
        }

        private void drawRobot()
        {
            Bitmap bmpRobot = new Bitmap(pictureBoxRobot.Width, pictureBoxRobot.Height);
            Graphics g = Graphics.FromImage(bmpRobot);

            int width = bmpRobot.Width - 1;
            int height = bmpRobot.Height - 1;


            g.Clear(Color.Black);       //background is transparent
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            GraphicsPath path = new GraphicsPath();
            SolidBrush pathBrush = new SolidBrush(Color.Blue);
            Pen drawPen = new Pen(pathBrush);
            drawPen.Color = Color.Yellow;
            int r = 10;
            g.FillEllipse(pathBrush, width / 2 - r, height / 2 - r, 2 * r, 2 * r);

            int x = (int)(width / 2 - (double)(r + 5) * Math.Sin(robotPosition.robotDir));
            int y = (int)(height / 2 - (double)(r + 5) * Math.Cos(robotPosition.robotDir));

            g.DrawLine(drawPen, width / 2, height / 2, x, y);
            pictureBoxRobot.Image = bmpRobot;
            //move robot picture box on the map to show the position of robot
            int MapY = MapY0 - (int)(robotPosition.robotX / MapRatio * (double)VRes / 0.0254);
            int MapX = MapX0 - (int)(robotPosition.robotY / MapRatio * (double)HRes / 0.0254);

            MapX = (MapX < 10? 10:MapX);
            MapX = (MapX > pictureBoxMap.Width? pictureBoxMap.Width:MapX);
            
            MapY = (MapY < 10? 10:MapY);
            MapY = (MapY > pictureBoxMap.Height? pictureBoxMap.Height:MapY);

            pictureBoxRobot.Location = new Point(pictureBoxMap.Location.X + MapX - 10, pictureBoxMap.Location.Y + MapY - 10);

        }

        private void wanderingNextPoint()
        {
            PointSet _wanderPointSet = new PointSet();
            PointSet.PointConfigTableRow row = (PointSet.PointConfigTableRow)_wanderPointSet.PointConfigTable.NewRow();
            //set target at the 1m away with the joystick direction
            row.TargetX = Double.Parse((robotPosition.robotX + Math.Cos(robotPosition.robotDir)).ToString("0.00"));
            row.TargetY = Double.Parse((robotPosition.robotY + Math.Sin(robotPosition.robotDir)).ToString("0.00"));
            row.TargetDir = Double.Parse((robotPosition.robotDir * 180 / Math.PI).ToString("0.00"));

            row.ForwardSpeed = trackBarWanderSpeed.Value / 100;
            row.ForwardSpeed = Math.Min(MAXSPEED, row.ForwardSpeed);
            row.ForwardSpeed = Math.Max(MINSPEED, row.ForwardSpeed);
            row.CAEnable = true;

            row.FinalPosture = false;
            row.Forgetable = true;
            row.MaxTurnSpeed = 35;      //35 degree
            row.NonStop = true;
            row.ReverseDrive = false;

            row.StopTime = 0;
            row.TargetTime = 200;
            row.TargetTolerance = 0.1;
            row.TargetDirTolerance = 15;        //15 degree

            _wanderPointSet.PointConfigTable.AddPointConfigTableRow(row);
            dataGridViewPointConfig.DataSource = _wanderPointSet.PointConfigTable;
            SendP2PGoCmd(0);
            

        }

        private void SendP2PGoCmd(int index)
        {
            DrRobotP2PSpeedDrive.SetTargetPointRequest request = new DrRobotP2PSpeedDrive.SetTargetPointRequest();

            if (dataGridViewPointConfig.Rows[index] != null)
            {
                request.TargetX = (double)dataGridViewPointConfig.Rows[index].Cells[0].Value;
                request.TargetY = (double)dataGridViewPointConfig.Rows[index].Cells[1].Value;
                request.TargetDir = (double)dataGridViewPointConfig.Rows[index].Cells[2].Value * Math.PI / 180;  //to rad
                request.StopTime = (int)dataGridViewPointConfig.Rows[index].Cells[3].Value;
                request.ForwardSpeed = (double)dataGridViewPointConfig.Rows[index].Cells[4].Value;
                request.Forgetable = (bool)dataGridViewPointConfig.Rows[index].Cells[5].Value;
                request.NonStop = (bool)dataGridViewPointConfig.Rows[index].Cells[6].Value;
                request.FinalPosture = (bool)dataGridViewPointConfig.Rows[index].Cells[7].Value;
                request.TargetTime = (int)dataGridViewPointConfig.Rows[index].Cells[8].Value;
                request.TargetTolerance = (double)dataGridViewPointConfig.Rows[index].Cells[9].Value;
                request.MaxTurnSpeed = (double)dataGridViewPointConfig.Rows[index].Cells[10].Value * Math.PI / 180; //to rad
                request.CAEnable = (bool)dataGridViewPointConfig.Rows[index].Cells[11].Value;
                request.ReverseDrive = (bool)dataGridViewPointConfig.Rows[index].Cells[12].Value;
                request.TargetDirTolerance = (double)dataGridViewPointConfig.Rows[index].Cells[13].Value * Math.PI / 180;  // to rad

            }
            p2pDrive.SetTargetPosition(request);
            p2pDrive.SendP2PCmd(DrRobotP2PSpeedDrive.P2PCtrlCmd.P2PGo);
            robotP2PStatus = DrRobotP2PSpeedDrive.P2PServiceStatus.P2PGo;

        }

        private void SendP2PStopCmd()
        {
            p2pDrive.SendP2PCmd(DrRobotP2PSpeedDrive.P2PCtrlCmd.StopP2P);
        }

        private void StucKHeatDetect()
        {

            //left first
            if (leftMotor .current  <= MotorHeatCurrentTH)
            {
                leftMotor.heatUpAcc = ( (leftMotor.heatUpAcc - (MotorHeatCurrentTH - leftMotor.current) * 0.5) > 0 ?
                    (leftMotor.heatUpAcc - (MotorHeatCurrentTH - leftMotor.current) * 0.5) : 0);
            }
            else
            {
                leftMotor.heatUpAcc = leftMotor.heatUpAcc + (leftMotor.current - MotorHeatCurrentTH) * 1.5;
            }

            if (leftMotor.heatUpAcc >= MotorHeatAccTH)
            {
                leftMotor.heatUpAcc = MotorHeatAccTH;
                leftMotor.heatUpProtect = true;
            }
            else
            {
                //take some time to heat gone away
                if ((leftMotor.heatUpProtect) && (leftMotor.heatUpAcc == 0))
                {
                    leftMotor.heatUpProtect = false;
                }

            }

            ////right
            if (rightMotor .current  <= MotorHeatCurrentTH)
            {
                rightMotor.heatUpAcc = ( (rightMotor.heatUpAcc - (MotorHeatCurrentTH - rightMotor.current) * 0.5) > 0 ?
                    (rightMotor.heatUpAcc - (MotorHeatCurrentTH - rightMotor.current) * 0.5) : 0);
            }
            else
            {
                rightMotor.heatUpAcc = rightMotor.heatUpAcc + (rightMotor.current - MotorHeatCurrentTH) * 1.5;
            }

            if (rightMotor.heatUpAcc >= MotorHeatAccTH)
            {
                rightMotor.heatUpAcc = MotorHeatAccTH;
                rightMotor.heatUpProtect = true;
            }
            else
            {
                //take some time to heat gone away
                if ((rightMotor.heatUpProtect) && (rightMotor.heatUpAcc == 0))
                {
                    rightMotor.heatUpProtect = false;
                }

            }


            // stuck detect
            leftMotor.stuckSpeedRatio = 0.5 * leftMotor.stuckSpeedRatio
                + 0.5 * Math.Abs(leftMotor.current/ (double)(Math.Abs(leftMotor.encoderSpeed) + 1));

            rightMotor.stuckSpeedRatio = 0.5 * rightMotor.stuckSpeedRatio
                + 0.5 * Math.Abs(rightMotor.current / (double)(Math.Abs(rightMotor.encoderSpeed) + 1));


           leftMotor.stuckLevel = (leftMotor.stuckSpeedRatio > StuckCurrentTH ?
                leftMotor .stuckLevel + StuckLevelUpStep : leftMotor .stuckLevel - StuckLevelDownStep);

           rightMotor.stuckLevel = (rightMotor.stuckSpeedRatio > StuckCurrentTH ?
               rightMotor.stuckLevel + StuckLevelUpStep : rightMotor.stuckLevel - StuckLevelDownStep);



           leftMotor.stuckLevel = (leftMotor .stuckLevel > MaxStuckLevel ? MaxStuckLevel + 1 : leftMotor .stuckLevel);
           rightMotor.stuckLevel = (rightMotor.stuckLevel > MaxStuckLevel ? MaxStuckLevel + 1 : rightMotor.stuckLevel);

           leftMotor.stuckLevel = (leftMotor.stuckLevel < 1 ? 0 : leftMotor.stuckLevel);
           rightMotor.stuckLevel = (rightMotor .stuckLevel < 1 ? 0 : rightMotor .stuckLevel);

           if (leftMotor.stuckLevel > StuckLevelTH)
            {
                leftMotor .stuckProtect = true;
            }
            else
            {
                if ((leftMotor .stuckProtect == true) && (leftMotor .stuckLevel < StuckLevelResetTH))
                {
                    leftMotor .stuckProtect = false;
                }

            }

           if (rightMotor.stuckLevel > StuckLevelTH)
           {
               rightMotor.stuckProtect = true;
           }
           else
           {
               if ((rightMotor.stuckProtect == true) && (rightMotor.stuckLevel < StuckLevelResetTH))
               {
                   rightMotor.stuckProtect = false;
               }

           }

           lblLeftMotorHeatProtectState.Text = leftMotor.heatUpProtect.ToString();
           lblRightMotorHeatProtectState.Text = rightMotor.heatUpProtect.ToString();

           lblLeftMotorStuckState.Text = leftMotor.stuckProtect.ToString();
           lblRightMotorStuckState.Text = rightMotor.stuckProtect.ToString();

        }

        private double AD2Dis(Int16 IRValue)
        {
            double temp = 0;
            double IRAD2Distance = 0;
            temp = 21.6 / ((double)IRValue * 3 / 4028 - 0.17);

            // IR range 10-80cm
            if ((temp > 80) || (temp < 0))
                IRAD2Distance = 0.81;
            else if ((temp < 10) && (temp > 0))
                IRAD2Distance = 0.09;
            else
                IRAD2Distance = temp / 100;

            return IRAD2Distance;

        }


        #endregion

        private void groupBox15_Enter(object sender, EventArgs e)
        {

        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start("http://www.drrobot.com");
        }
    }
}
